#include <stdio.h>
#include <stdlib.h>
#include <i86.h>
#include <dos.h>
#include <string.h>

// default settings
unsigned __based(__segname("_CODE")) screen_timeout = 300;
unsigned __based(__segname("_CODE")) hdd_timeout = 600;
unsigned __based(__segname("_CODE")) suspend_timeout = 1000;

unsigned __based(__segname("_CODE")) clock_ticks = 0;
unsigned __based(__segname("_CODE")) seconds = 0;

unsigned __based(__segname("_CODE")) api_version;

unsigned char __based(__segname("_CODE")) power_status = 0;

#define TICKS_PER_SECOND 18

unsigned current_psp(void);
#pragma aux current_psp = \
  "mov ah, 0x51" \
  "int 0x21" \
  value [bx] modify [ax];

int dos_close(unsigned handle);
#pragma aux dos_close = \
  "mov ah, 0x3e" \
  "int 0x21" \
  parm [bx] value [ax];

void __far dos_free(unsigned seg);
#pragma aux dos_free = \
  "mov ah, 0x49" \
  "mov es, bx" \
  "int 0x21" \
  parm [bx] modify [es ax];

int apm_call(unsigned char call, unsigned device, unsigned code);
#pragma aux apm_call = \
  "mov ah, 0x53" \
  "int 0x15" \
  "jc end" \
  "xor ax, ax" \
  "end:" \
  parm [al] [bx] [cx] value [ax];

int apm_version(void);
#pragma aux apm_version = \
  "mov ax, 0x530E" \
  "xor bx, bx" \
  "mov cx, 0x0102" \
  "int 0x15" \
  "jc error" \
  "jmp end" \
  "error:" \
  "mov ax, 0x0100" \
  "end:" \
  modify exact [ax bx cx] value [ax];

inline void hdd_power_off(void)
{
  apm_call(0x07, 0x02FF, 3);
  power_status |= 2;
}

void hdd_power_on(void)
{
  apm_call(0x07, 0x02FF, 0);
  power_status &= ~2;
}

void display_power_off(void)
{
  apm_call(0x07, 0x01FF, 3);
  power_status |= 1;
}

void display_power_on(void)
{
  apm_call(0x07, 0x01FF, 0);
  power_status &= ~1;
}

void cpu_idle(void);
#pragma aux cpu_idle = \
  "mov ax, 0x5305" \
  "int 0x15" \
  modify exact [ax];

void cpu_busy(void);
#pragma aux cpu_busy = \
  "mov ax, 0x5306" \
  "int 0x15"  \
  modify exact [ax];


void throttle_cpu(void)
{
  cpu_idle();
  power_status |= 4;
}

void restore_cpu(void)
{
  cpu_busy();
  power_status &= ~4;
}

void halt(void);
#pragma aux halt = \
  "hlt" \
  modify exact [];

void restore_power(void)
{
  if(power_status & 1)
    display_power_on();
  if(power_status & 2)
    hdd_power_on();
}

void suspend_system(void)
{
  apm_call(0x07, 0x0001, 2);
// TODO: set interrupt mask to keyboard only
  seconds = 0;
  restore_power();
}

void flush_buffers(void);
#pragma aux flush_buffers = \
  "mov ah, 0x0D" \
  "int 0x21" \
  modify [ax];

void dos_outputch(char);
#pragma aux dos_outputch = \
  "mov ah, 0x02" \
  "int 0x21" \
  parm [dl] modify [ax];

char __based(__segname("_CODE")) shutdown_str[] = \
"All file buffers have been flushed to the disk. It is now safe to turn off \
your computer.";

void shutdown_system(void)
{
  unsigned n;
  flush_buffers();
  for(n=0;shutdown_str[n];n++)
    dos_outputch(shutdown_str[n]);
  apm_call(0x07, 0x0001, 3);
}

int apm_connect(void);
#pragma aux apm_connect = \
  "mov ax, 0x5301" \
  "xor bx, bx" \
  "int 0x15" \
  "jc end" \
  "xor ax, ax" \
  "end:" \
  value [ax] modify exact [ax bx];

void apm_disconnect(void);
#pragma aux apm_disconnect = \
  "mov ax, 0x5304" \
  "xor bx, bx" \
  "int 0x15" \
  modify exact [ax bx];

void timer_routine(void)
{
  if(++clock_ticks == TICKS_PER_SECOND)
  {
    clock_ticks = 0;
    seconds++;
  }
  if(screen_timeout && seconds > screen_timeout && !(power_status & 1))
    display_power_off();
  if(hdd_timeout && seconds > hdd_timeout && !(power_status & 2))
    hdd_power_off();
  if(suspend_timeout && seconds > suspend_timeout)
    suspend_system();
}


extern void __far * __based(__segname("_CODE")) old_timer_rtn;
extern void __far * __based(__segname("_CODE")) old_int2f_handler;
void __far timer_int(void);
void __far int2f_handler(void);
void __far int13_handler(void);
extern void __far * __based(__segname("_CODE")) old_int13_handler;

extern void __far * __based(__segname("_CODE")) old_irq1_handler;
extern void __far * __based(__segname("_CODE")) old_irq3_handler;
extern void __far * __based(__segname("_CODE")) old_irq4_handler;
extern void __far * __based(__segname("_CODE")) old_irq5_handler;
extern void __far * __based(__segname("_CODE")) old_irq6_handler;
extern void __far * __based(__segname("_CODE")) old_irq7_handler;

extern void __far * __based(__segname("_CODE")) old_irq9_handler;
extern void __far * __based(__segname("_CODE")) old_irq10_handler;
extern void __far * __based(__segname("_CODE")) old_irq11_handler;
extern void __far * __based(__segname("_CODE")) old_irq12_handler;
extern void __far * __based(__segname("_CODE")) old_irq14_handler;
extern void __far * __based(__segname("_CODE")) old_irq15_handler;

extern __far irq1_handler(void);
extern __far irq3_handler(void);
extern __far irq4_handler(void);
extern __far irq5_handler(void);
extern __far irq6_handler(void);
extern __far irq7_handler(void);

extern __far irq9_handler(void);
extern __far irq10_handler(void);
extern __far irq11_handler(void);
extern __far irq12_handler(void);
extern __far irq14_handler(void);
extern __far irq15_handler(void);

int apm_installation_check(void);
#pragma aux apm_installation_check = \
  "mov ax, 0x5300" \
  "xor bx, bx" \
  "int 0x15" \
  "jc error" \
  "cmp bx, 0x504D" \
  "jne error" \
  "xor ax, ax" \
  "jmp end" \
  "error:" \
  "end:" \
  modify exact [ax bx cx] value [ax];

void __far * interrupt_handlers_low[] =
{
  0, 
  irq1_handler,
  0, 
  irq3_handler,
  irq4_handler,
  irq5_handler,
  irq6_handler,
  irq7_handler
};

void __far * interrupt_handlers_high[] =
{
  0,
  irq9_handler,
  irq10_handler,
  irq11_handler,
  irq12_handler,
  0,
  irq14_handler,
  irq15_handler
};

void zero_counters_on_irq(unsigned irq)
{
  if(irq < 8)
    _dos_setvect(irq+8, interrupt_handlers_low[irq]);
  else
    _dos_setvect(irq+70, interrupt_handlers_high[irq-8]);
}

int load_settings(int argc, char *argv[])
{
  const char display_timeout_str[] = "DISPLAY_TIMEOUT=";
  const char hdd_timeout_str[] = "HDD_TIMEOUT=";
  const char suspend_timeout_str[] = "SUSPEND_TIMEOUT=";
  const char wakeup_irq_str[] = "WAKEUP_ON_IRQ=";
  char buff[100];
  unsigned n;
  FILE *fp;

  for(n=1;n<argc;n++)
    if(strcmp(argv[n], "/F"))
    {
      fp = fopen(argv[n], "r");
      goto fp_opened;
    }
  fp = fopen("power.ini", "r");
  fp_opened:

  if(!fp)
  {
    puts("Cannot open POWER.INI, using default settings...");
    zero_counters_on_irq(1);
    return 1;
  }

  while(fgets(buff, sizeof(buff), fp))
  {
    if(!memcmp(wakeup_irq_str, buff, sizeof(wakeup_irq_str)-1))
    {
      n = atoi(&buff[sizeof(wakeup_irq_str)-1]);
      zero_counters_on_irq(n);
      printf("Zeroing counters on IRQ %u.\n", n);
    }
    if(!memcmp(display_timeout_str, buff, sizeof(display_timeout_str)-1))
    {
      screen_timeout = atoi(&buff[sizeof(display_timeout_str)-1]);
      printf("Power off display after %u seconds\n", screen_timeout);
    }
    else if(!memcmp(hdd_timeout_str, buff, sizeof(hdd_timeout_str)-1))
    {
      hdd_timeout = atoi(&buff[sizeof(hdd_timeout_str)-1]);
      printf("Power off hard disks after %u seconds\n", hdd_timeout);
    }
    else if(!memcmp(suspend_timeout_str, buff, sizeof(suspend_timeout_str)-1))
    {
      suspend_timeout = atoi(&buff[sizeof(suspend_timeout_str)-1]);
      printf("Suspend mode after %u seconds\n", suspend_timeout);
    }
  }

  fclose(fp);
  return 0;
}


int main(int argc, char *argv[])
{
  unsigned env_seg;
  unsigned char __far *psp_ptr = MK_FP(current_psp(), 0);
  unsigned n;
  unsigned force;

  if(argc > 1)
    force = !strcmp(argv[1], "/F");
  else force=0;

  puts("Loading APM power saving driver v. 1.0.");



  if(apm_installation_check() && !force)
  {
    puts("Error: APM is not supported by your BIOS.");
    return 1;
  }

  if(apm_connect() && !force)
  {
    puts("APM interface is already connected.");
    return 1;
  }

  api_version = apm_version();
  printf("BIOS APM version: %.4X\n", api_version);

  if(api_version == 0x0100)
    apm_call(0x08, 0xFFFF, 0x0001);
  else
    apm_call(0x08, 0x0001, 0x0001);

  // engage cooperative power management
  apm_call(0x0F, 0x0001, 0x0001);
  // enable automatic power management
  apm_call(0x0D, 0x0001, 0x0001);

  old_timer_rtn = _dos_getvect(0x1C);
  _dos_setvect(0x1C, timer_int);
  old_int2f_handler = _dos_getvect(0x2F);
  _dos_setvect(0x2F, int2f_handler);
  old_int13_handler = _dos_getvect(0x13);
  _dos_setvect(0x13, int13_handler);


  old_irq1_handler = _dos_getvect(9);
  old_irq3_handler = _dos_getvect(0xB);
  old_irq4_handler = _dos_getvect(0xC);
  old_irq5_handler = _dos_getvect(0xD);
  old_irq6_handler = _dos_getvect(0xE);
  old_irq7_handler = _dos_getvect(0xf);
  old_irq9_handler = _dos_getvect(71);
  old_irq10_handler = _dos_getvect(72);
  old_irq11_handler = _dos_getvect(73);
  old_irq12_handler = _dos_getvect(74);
  old_irq14_handler = _dos_getvect(76);
  old_irq15_handler = _dos_getvect(77);

  load_settings(argc, argv);

  env_seg = *(unsigned __far*)&psp_ptr[0x2C];
  dos_free(env_seg);

  for(n=3;n--;)
    dos_close(n);

  _dos_keep(0, 0x10 + ((FP_OFF(zero_counters_on_irq) + 0xF) >> 4));

  puts("Error");
  return 1;
}
