.8086

.MODEL small

extern timer_routine_:BYTE
extern _seconds:WORD
extern restore_power_:BYTE
extern hdd_power_on_:BYTE
extern _power_status:BYTE
extern throttle_cpu_:BYTE
extern restore_cpu_:BYTE
extern display_power_off_:BYTE
extern display_power_on_:BYTE
extern suspend_system_:BYTE
extern shutdown_system_:BYTE
extern _screen_timeout:WORD
extern _hdd_timeout:WORD
extern _suspend_timeout:WORD
public _old_timer_rtn
public timer_int_
public _old_int2f_handler
public int2f_handler_
public int13_handler_
public _old_int13_handler

public irq1_handler_
public _old_irq1_handler
public irq3_handler_
public _old_irq3_handler
public irq4_handler_
public _old_irq4_handler
public irq5_handler_
public _old_irq5_handler
public irq6_handler_
public _old_irq6_handler
public irq7_handler_
public _old_irq7_handler
public irq9_handler_
public _old_irq9_handler
public irq10_handler_
public _old_irq10_handler
public irq11_handler_
public _old_irq11_handler
public irq12_handler_
public _old_irq12_handler
public irq14_handler_
public _old_irq14_handler
public irq15_handler_
public _old_irq15_handler

_TEXT               SEGMENT BYTE PUBLIC USE16 'CODE'
                ASSUME CS:_TEXT                               

timer_int_ proc far
  push ax
  push es

  call near ptr timer_routine_

  pop es
  pop ax

  db 0xEA
  _old_timer_rtn: dd 0
timer_int_ endp

; todo: services:
; - installation check
; - zero counters
; - throttle cpu clock
; - restore cpu clock
int2f_handler_ proc far
  cmp ah, 0x54
  jne next_int2f_handler

; installation check
  cmp al, 0x00
  jnz int2f_jmp_1
  cmp bx, 0x504D
  jne int2f_end
  mov bx, 0x5354
  mov ax, 0x0100
  iret

; zero counters
  int2f_jmp_1:
  cmp al, 0x01
  jnz int2f_jmp_2
  mov cs:_seconds, 0
  iret

; throttle CPU clock (destroys ax, es)
  int2f_jmp_2:
  cmp al, 0x02
  jnz int2f_jmp_3
  call near ptr throttle_cpu_
  iret

; restore CPU clock (destroys ax, es)
  int2f_jmp_3:
  cmp al, 0x03
  jnz int2f_jmp_4
  call near ptr restore_cpu_
  iret

; shut down displays manually (destroys ax, es)
  int2f_jmp_4:
  cmp al, 0x04
  jnz int2f_jmp_5
  call near ptr display_power_off_
  iret

; power on displays (destroys ax, es)
  int2f_jmp_5:
  cmp al, 0x05
  jnz int2f_jmp_6
  call near ptr display_power_on_
  iret

  int2f_jmp_6:
; suspend the system (destroys ax, es)
  cmp al, 0x06
  jnz int2f_jmp_7
  call near ptr suspend_system_
  iret

; shut down the system
  int2f_jmp_7:
  cmp al, 0x07
  jnz int2f_jmp_8
  call near ptr shutdown_system_
  cli
  hlt

; change display shutdown timeout (destroys ax, es)
  int2f_jmp_8:
  cmp al, 0x08
  jnz int2f_jmp_9
  xchg cs:_screen_timeout, bx
  iret

; change hdd timeout (destroys ax, es)
  int2f_jmp_9:
  cmp al, 0x09
  jnz int2f_jmp_10
  xchg cs:_hdd_timeout, bx
  iret

; change suspend timeout (destroys ax, es)
  int2f_jmp_10:
  cmp al, 0x0A
  jnz  int2f_jmp_11
  xchg cs:_suspend_timeout, bx
  iret

  int2f_jmp_11:
  int2f_end:
  iret

  next_int2f_handler:
  db 0xEA
  _old_int2f_handler: dd 0
int2f_handler_ endp

int13_handler_ proc far
  test cs:_power_status, 2
  jz int13_handler_end
  mov cs:_seconds, 0

  push ax
  push es
  call near ptr hdd_power_on_
  pop es
  pop ax
  
  int13_handler_end:
  db 0xEA
  _old_int13_handler: dd 0
int13_handler_ endp

zero_counters:
  mov cs:_seconds, 0
  push ax
  push es
  call near ptr restore_power_
  pop es
  pop ax
ret

; keyboard irq
irq1_handler_ proc far
  call zero_counters
  db 0xEA
  _old_irq1_handler: dd 0
irq1_handler_ endp

; COM2/COM4
irq3_handler_ proc far
  call zero_counters
  db 0xEA
  _old_irq3_handler: dd 0
irq3_handler_ endp

; COM1/COM3
irq4_handler_ proc far
  call zero_counters
  db 0xEA
  _old_irq4_handler: dd 0
irq4_handler_ endp

; LPT2
irq5_handler_ proc far
  call zero_counters
  db 0xEA
  _old_irq5_handler: dd 0
irq5_handler_ endp

; floppy
irq6_handler_ proc far
  call zero_counters
  db 0xEA
  _old_irq6_handler: dd 0
irq6_handler_ endp

; LPT1
irq7_handler_ proc far
  call zero_counters
  db 0xEA
  _old_irq7_handler: dd 0
irq7_handler_ endp

; often a LAN card
irq9_handler_ proc far
  call zero_counters
  db 0xEA
  _old_irq9_handler: dd 0
irq9_handler_ endp

irq10_handler_ proc far
  call zero_counters
  db 0xEA
  _old_irq10_handler: dd 0
irq10_handler_ endp

; reserved
irq11_handler_ proc far
  call zero_counters
  db 0xEA
  _old_irq11_handler: dd 0
irq11_handler_ endp

; mouse
irq12_handler_ proc far
  call zero_counters
  db 0xEA
  _old_irq12_handler: dd 0
irq12_handler_ endp

; primary IDE
irq14_handler_ proc far
  call zero_counters
  db 0xEA
  _old_irq14_handler: dd 0
irq14_handler_ endp

;secondary IDE
irq15_handler_ proc far
  call zero_counters
  db 0xEA
  _old_irq15_handler: dd 0
irq15_handler_ endp

_TEXT ENDS

END
